#ifndef OATH_BIGINTEGER_H

#define OATH_BIGINTEGER_H

//***************************************************************************
//             OATH :: Object-oriented Abstract Type Hierarchy
//***************************************************************************
//
//  Copyright (C) 1991  Texas Instruments Incorporated
//  Permission is granted to any individual or institution
//  to use, copy, modify, and distribute this software,
//  provided that this complete copyright and permission notice
//  is maintained, intact, in all copies and supporting documentation.
//
//  Texas Instruments Incorporated provides this software "as is"
//  without express or implied warranty.
//
//***************************************************************************
//  bigInteger (bigIntegerA, bigIntegerG)
//
//  History:
//    06/92  Brian M Kennedy  restored ~bigIntegerG which had disappeared
//	                      changed 0U's to 0L's
//    08/91  Brian M Kennedy  signMagnitudeP; functionality refined
//    07/91  Brian M Kennedy  import, export, typeRegister
//    06/91  Brian M Kennedy  New macros & format; remove printDiagnostic
//    02/91  Brian M Kennedy  Original
//
//***************************************************************************

#include "numeric.h"

#include "signMagnitudeP.h"

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////
// Class Definitions

/////////////////////////////////////////////////////////////////////////////
// class bigIntegerA

CLASS_A(bigInteger, integer)

        // Makes ////////////////////

	    static inline bigIntegerA
	make (long L = 0, int MakeConst = FALSE);

	    static inline bigIntegerA
	make (const complexA& C, int MakeConst = FALSE);

	    static inline bigIntegerA
	make (const char* S, int MakeConst = FALSE);

        // Operations ////////////////////


	integer_RETURNING_DEFINITIONS(bigInteger)

   };


/////////////////////////////////////////////////////////////////////////////
// class bigIntegerG

CLASS_G(bigInteger, integer)

	// Data Members //////////

	unsigned int   MaxSize;
	signMagnitudeP SAM;

	// Static Data Members //////////

	static unsigned long  ZeroSAMMagnitude;
	static signMagnitudeP ZeroSAM;

	// Constructors //////////

	    inline
        bigIntegerG (long IValue = 0,            int IsConst = FALSE);

        bigIntegerG (const char* IValue,         int IsConst = FALSE);

        bigIntegerG (const bigIntegerG* IValue,  int IsConst = FALSE);

	bigIntegerG (const signMagnitudeP& ISAM, int IsConst = FALSE);

	    inline
	bigIntegerG (unsigned int IMaxSize, unsigned int ISAS,
	             unsigned long * IMag,  int IsConst = FALSE);

	    inline
        ~bigIntegerG ();

        // oathCore Operations //////////

	    virtual inline int
	isImplementedAs (void* TypeId) const;

	    virtual void
	export (exportP& X) const;

	    static objA
	import (importP& M);


        // obj Operations //////////

	    virtual int
	isEqual (const objG* O) const;

	    virtual inline objA
	makeCopy (int MakeConst) const;


        // complex Operations //////////

	    virtual void
	reset ();

	    virtual void
	reset (const complexG* C);

	    virtual inline realA
	makeReal (int MakeConst) const;

	    virtual inline realA
	makeImag (int MakeConst) const;

	    virtual inline realA
	makeAbs (int MakeConst) const;

	    virtual inline realA
	makeArg (int MakeConst) const;

	    virtual inline int
	hasZero () const;

	    virtual inline int
	hasZeroImag () const;

	    virtual inline int
	hasEqual (const complexG* C) const;

	    virtual inline void
	increment ();

	    virtual void
	increment (const complexG* C);

	    virtual inline void
	decrement ();

	    virtual void
	decrement (const complexG* C);

	    virtual void
	multiply (const complexG* C);

	    virtual void
	divide (const complexG* C);

	    virtual void
	shiftLeft (const integerG* Count);

	    virtual void
	shiftRight (const integerG* Count);

	    virtual inline void
	neg ();

	    virtual inline void
	conj ();

	    virtual inline void
	abs ();

	    virtual inline void
	inv ();

	    virtual void
	sqr ();

	    virtual void
	sqrt ();

	    virtual void
	pow (const integerG* Expn);

	    virtual void
	print (ostream& F, char Sep) const;


        // real Operations //////////

	    virtual inline const signMagnitudeP&
	smMantissa () const;

	    virtual inline const signMagnitudeP&
	smExponent () const;

	    virtual inline integerA
	makeFloor (int MakeConst) const;

	    virtual inline integerA
	makeCeiling (int MakeConst) const;

	    virtual inline integerA
	makeRound (int MakeConst) const;

	    virtual inline integerA
	makeTruncate (int MakeConst) const;

	    virtual inline int
	sign () const;

	    virtual inline int
	hasPositive () const;

	    virtual inline int
	hasNegative () const;

	    virtual int
	hasLess (const realG* R) const;

	    virtual int
	hasMore (const realG* R) const;

	    virtual int
	hasDouble () const;

	    virtual double
	makeDouble () const;

	    virtual void
	min (const realG* R);

	    virtual void
	max (const realG* R);


        // rational Operations //////////

	    virtual inline integerA
	makeNumerator (int MakeConst) const;

	    virtual inline integerA
	makeDenominator (int MakeConst) const;


        // integer Operations //////////

	    virtual inline int
	hasEven () const;

	    virtual inline int
	hasLong () const;

	    virtual inline long
	makeLong () const;

	    virtual void
	rem (const integerG*);

	    virtual void
	remDivX (const integerG*, integerG*);

	    virtual void
	gcd (const integerG* I);

	    virtual void
	lcm (const integerG* I);


        // bigInteger Operations //////////

	    void
	smGrowMag (unsigned int NewMax);

	    void
	smReset (long L);

	    void
	smReset (const signMagnitudeP& SMI);

	// Shifts

	    void
	smShiftLeft (unsigned int Count);

	    void
	smShiftLeft (const signMagnitudeP& SMI);

	    void
	smShiftRight (unsigned int Count);

	    void
	smShiftRight (const signMagnitudeP& SMI);

	// Comparison

	    int
	smHasEqualMag (const signMagnitudeP& SMI) const;

	    int
	smHasEqual (const signMagnitudeP& SMI) const;

	    int
	smHasLessMag (const signMagnitudeP& SMI) const;

	    int
	smHasMoreMag (const signMagnitudeP& SMI) const;

	    int
	smHasLess (const signMagnitudeP& SMI) const;

	    int
	smHasMore (const signMagnitudeP& SMI) const;

	    void
	smMin (const signMagnitudeP& SMI);

	    void
	smMax (const signMagnitudeP& SMI);

	// Increment

	    void
	smIncExtend ();

	    void
	smIncMag ();

	    void
	smIncMag (unsigned long UL);

	    void
	smIncMag (const signMagnitudeP& SMI);

	    void
	smInc (const signMagnitudeP& SMI);

	// Decrement

	    void
	smDecMag ();

	    void
	smDecMag (unsigned long UL);

	    void
	smDecMag (const signMagnitudeP& SMI);

	    void
	smDec (const signMagnitudeP& SMI);

	// Multiplication

	    inline void
	smMul10 ();

	    void
	smMulMag (const signMagnitudeP& SMI);

	    void
	smMul (const signMagnitudeP& SMI);

	// Division

	    unsigned long
	smDivRem10 ();

	    void
	smRemDivMag (const signMagnitudeP& Dsor,
		     signMagnitudeP& Quot, unsigned int& QuotMax);

	    void
	smDiv (const signMagnitudeP& SMI);

	    void
	smMod (const signMagnitudeP& SMI);

	    void
	smRem (const signMagnitudeP& SMI);

	// Output

	    char *
	string (char* Buffer, int N, char Sep = '\0') const;

   };


/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////
// Inline Definitions

/////////////////////////////////////////////////////////////////////////////
// bigIntegerG Inlines

// Constructors //////////

    inline bigIntegerG::
bigIntegerG (long IValue, int IsConst)
   :integerG(IsConst), MaxSize(1),
    SAM(1, (IValue >= 0), new unsigned long [1])
   {SAM.magnitude(0) = IValue >= 0 ? IValue : -IValue;}

    inline bigIntegerG::
bigIntegerG (unsigned int IMaxSize, unsigned int ISAS,
	     unsigned long * IMag,  int IsConst)
   :integerG(IsConst), MaxSize(IMaxSize), SAM(ISAS, IMag)
   {}

    inline bigIntegerG::
~bigIntegerG ()
   {delete [] SAM.magnitude();}


// oathCore Operations //////////

    inline int bigIntegerG::
isImplementedAs (void* TypeId) const
   {return TypeId == Type;}


// obj Operations //////////

    inline objA bigIntegerG::
makeCopy (int MakeConst) const
   {return new bigIntegerG (SAM, MakeConst);}


// complex Operations //////////

    inline realA bigIntegerG::
makeReal (int MakeConst) const
   {return new bigIntegerG (this, MakeConst);}

    inline realA bigIntegerG::
makeImag (int MakeConst) const
   {return new bigIntegerG (0L, MakeConst);}

    inline int bigIntegerG::
hasZero () const
   {return SAM.zero();}

    inline int bigIntegerG::
hasZeroImag () const
   {return TRUE;}

    inline void bigIntegerG::
neg ()
   {SAM.negate();}


// real Operations //////////

    inline int bigIntegerG::
hasPositive () const
   {return SAM.positive() && !SAM.zero();}

    inline int bigIntegerG::
hasNegative () const
   {return !SAM.positive() && !SAM.zero();}

    inline int bigIntegerG::
sign () const
   {return SAM.zero() ? 0 : (SAM.positive()  ? 1 : -1);}


// complex Operations, continued //////////

    inline realA bigIntegerG::
makeAbs (int MakeConst) const
   {int N = hasNegative();
    if(N) ((bigIntegerG*)this)->neg();
    bigIntegerA I = new bigIntegerG (this, MakeConst);
    if(N) ((bigIntegerG*)this)->neg();
    return I;
   }

    inline realA bigIntegerG::
makeArg (int MakeConst) const
   {if(hasNegative())
        return new bigIntegerG (3, MakeConst);  // round pi
    else
        return new bigIntegerG (0L, MakeConst);
   }

    inline void bigIntegerG::
increment ()
   {if(hasPositive()) smIncMag();
    else	      smDecMag();
   }

    inline void bigIntegerG::
decrement ()
   {if(hasPositive()) smDecMag();
    else	      smIncMag();
   }

    inline void bigIntegerG::
conj ()
   {}

    inline void bigIntegerG::
abs ()
   {SAM.positive(1);}

    inline void bigIntegerG::	// whatever / does : trunc or round
inv ()
   {ensure(!SAM.zero(), "throw DivisionByZero::Overflow::RangeViolation;");
    if(SAM.size() == 1 && SAM.magnitude(0) == 1)
       {if(SAM.positive()) smReset(1);
        else		   smReset(-1);
       }
    else		   reset();
   }


// real Operations //////////

    inline const signMagnitudeP& bigIntegerG::
smMantissa () const
   {return SAM;}

    inline const signMagnitudeP& bigIntegerG::
smExponent () const
   {return ZeroSAM;}

    inline integerA bigIntegerG::
makeFloor (int MakeConst) const
   {return new bigIntegerG (this, MakeConst);}

    inline integerA bigIntegerG::
makeCeiling (int MakeConst) const
   {return new bigIntegerG (this, MakeConst);}

    inline integerA bigIntegerG::
makeRound (int MakeConst) const
   {return new bigIntegerG (this, MakeConst);}

    inline integerA bigIntegerG::
makeTruncate (int MakeConst) const
   {return new bigIntegerG (this, MakeConst);}


// rational Operations //////////

    inline integerA bigIntegerG::
makeNumerator (int MakeConst) const
   {return new bigIntegerG (this, MakeConst);}

    inline integerA bigIntegerG::
makeDenominator (int MakeConst) const
   {return new bigIntegerG (1, MakeConst);}


// integer Operations //////////

    inline int bigIntegerG::
hasEven () const
   {return !(SAM.magnitude(0) & 1UL);}

    inline int bigIntegerG::
hasLong () const
   {return (SAM.size() == 1) && (SAM.magnitude(0) < (SAM.magnitude(0) << 1));}

    inline long bigIntegerG::
makeLong () const
   {return SAM.positive() ? (long)(SAM.magnitude(0)) 
			  : -(long)(SAM.magnitude(0));
   }


// bigInteger Operations //////////

    inline void bigIntegerG::
smMul10 ()
   {if(!SAM.zero())
       {smShiftLeft(1);
        bigIntegerA D = (bigIntegerA&)makeCopy(TRUE);
	smShiftLeft(2);
	smIncMag(D.guts()->SAM);
       }
   }


/////////////////////////////////////////////////////////////////////////////
// bigIntegerA Inlines

INLINES(bigInteger, integer)

// Makes ////////////////////

    inline bigIntegerA bigIntegerA::
make (long L, int MakeConst)
   {return new bigIntegerG (L, MakeConst);}

    inline bigIntegerA bigIntegerA::
make (const char* S, int MakeConst)
   {return new bigIntegerG (S, MakeConst);}

// Operations ////////////////////


integer_RETURNING_INLINES(bigInteger)


//***************************************************************************

#endif /* OATH_BIGINTEGER_H */
