#ifndef OATH_CHARACTERSET_H

#define OATH_CHARACTERSET_H

//***************************************************************************
//             OATH :: Object-oriented Abstract Type Hierarchy
//***************************************************************************
//
//  Copyright (C) 1991, 1990  Texas Instruments Incorporated
//  Permission is granted to any individual or institution
//  to use, copy, modify, and distribute this software,
//  provided that this complete copyright and permission notice
//  is maintained, intact, in all copies and supporting documentation.
//
//  Texas Instruments Incorporated provides this software "as is"
//  without express or implied warranty.
//
//***************************************************************************
//  characterSet (characterSetA, characterSetG)
//
//  History:
//    07/91  Brian M Kennedy  import, export, typeRegister
//    06/91  Brian M Kennedy  New macros & format; remove printDiagnostic
//    04/91  Brian M Kennedy  Derived from finiteSet (was a set)
//    12/90  Brian M Kennedy  Original
//
//***************************************************************************

#include "finiteSet.h"

#include "character.h"

#include "minString.h"

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////
// Class Definitions

/////////////////////////////////////////////////////////////////////////////
// class characterSetA

CLASS_A(characterSet, finiteSet)

	// Makes //////////

	    static inline characterSetA
	make (int MakeConst = FALSE);

	    static inline characterSetA
	make (const stringA& S, int MakeConst = FALSE);

	    static inline characterSetA
	make (const char * S, int MakeConst = FALSE);


	// Operations //////////

	    inline const characterSetA&
        insert (unsigned char) const;

	    inline const characterSetA&
        operator << (unsigned char) const;

	    inline const characterSetA&
        insert (const characterA&) const;

	    inline const characterSetA&
        operator << (const characterA&) const;

	    inline int
        contains (unsigned char) const;

	    inline int
        contains (const characterA&) const;

	    inline const characterSetA&
        remove (unsigned char) const;

	    inline const characterSetA&
        operator -= (unsigned char) const;

	    inline const characterSetA&
        remove (const characterA&) const;

	    inline const characterSetA&
        operator -= (const characterA&) const;


	finiteSet_RETURNING_DEFINITIONS(characterSet)

   };


/////////////////////////////////////////////////////////////////////////////
// class characterSetG

CLASS_G(characterSet, finiteSet)

	// Static Members //////////

	static unsigned char BitMasks [8];

	// Members //////////

	unsigned char Bytes [32];

	// Constructors //////////

        characterSetG (int IsConst = FALSE);

	characterSetG (const unsigned char iBytes[32], int IsConst = FALSE);

	characterSetG (const stringG* S, int IsConst = FALSE);

    public:
        // characterSet Accessors //////////

	    inline int
	getbit (unsigned char C) const;

	    inline void
	clearbit (unsigned char C);

	    inline void
	setbit (unsigned char C);

	    void
	setbits (unsigned char S, unsigned char B);

	    void
	flipbits ();


	// oathCore Operations //////////

	    inline virtual int
	isImplementedAs (void* TypeId) const;

	    virtual void
	export (exportP& X) const;

	    static objA
	import (importP& M);


	// obj Operations //////////

	    virtual int
	isEqual (const objG* O) const;

	    inline virtual objA
	makeCopy (int IsConst) const;


        // bag Operations //////////

	    virtual int
	isEmpty () const;

	    virtual int
	count () const;

	    inline virtual int
	contains (const objG* O) const;

	    inline virtual int
	containsEqual (const objG* O) const;

	    inline virtual int
	canContain (const objG* O) const;

	    inline virtual void
	insert (const objG* O);

	    inline virtual void
	append (const bagG*);

	    virtual void
	apply (void (*)(objA)) const;

	    virtual bagG*
	applyX (objA (*)(objA), bagG* B) const;

	    virtual bagG*
	applyX (bagA (*)(objA), bagG* B) const;

	    inline virtual bagA
	makeEmpty () const;


        // set Operations //////////

	    inline virtual void
	remove (const objG* O);

	    virtual void
	subtract (const setG* S);

	    virtual void
	intersect (const setG* S);


        // finiteSet Operations //////////

	    virtual void
	invert ();

   };


/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////
// Inline Definitions

/////////////////////////////////////////////////////////////////////////////
// characterSetG Inlines

// characterSet Accessors //////////

    inline int characterSetG::
getbit (unsigned char C) const
   {return Bytes[C >> 3] & BitMasks[C & 7];}

    inline void characterSetG::
clearbit (unsigned char C)
   {Bytes[C >> 3] &= ~BitMasks[C & 7];}

    inline void characterSetG::
setbit (unsigned char C)
   {Bytes[C >> 3] |= BitMasks[C & 7];}


// oathCore Operations //////////

    inline int characterSetG::
isImplementedAs (void* TypeId) const
   {return Type == TypeId;}


// obj Operations //////////

    inline objA characterSetG::
makeCopy (int IsConst) const
   {return new characterSetG (Bytes, IsConst);}


// bag Operations //////////

    inline int characterSetG::
contains (const objG* O) const 
   {if(O->isImplementedAs(characterG::Type))
	return getbit(((const characterG*)O)->value());
    else
	return FALSE;
   }

    inline int characterSetG::
containsEqual (const objG* O) const 
   {if(O->isImplementedAs(characterG::Type))
	return getbit(((const characterG*)O)->value());
    else
	return FALSE;
   }

    inline int characterSetG::
canContain (const objG* O) const 
   {return O->isImplementedAs(characterG::Type);}

    inline void characterSetG::
insert (const objG* O)
   {if(O->isImplementedAs(characterG::Type))
	setbit(((const characterG*)O)->value());
   }

    inline void characterSetG::
append (const bagG*)
   {ensure(FALSE, "Sorry, not implemented!");}

    inline bagA characterSetG::
makeEmpty () const
   {return new characterSetG ();}


// set Operations //////////

    inline void characterSetG::
remove (const objG* O)
   {if(O->isImplementedAs(characterG::Type))
	clearbit(((const characterG*)O)->value());
   }


/////////////////////////////////////////////////////////////////////////////
// characterSetA Inlines

// Makes //////////

    inline characterSetA characterSetA::
make (int MakeConst)
   {return new characterSetG (MakeConst);}

    inline characterSetA characterSetA::
make (const stringA& S, int MakeConst)
   {return new characterSetG (S.guts(), MakeConst);}

    inline characterSetA characterSetA::
make (const char * S, int MakeConst)
   {return make(minStringA::make(S, TRUE), MakeConst);}

// Operations //////////

INLINES(characterSet, finiteSet)

    inline const characterSetA& characterSetA::
insert (unsigned char C) const
   {guts()->setbit(C);
    return *this;
   }

    inline const characterSetA& characterSetA::
operator << (unsigned char C) const
   {guts()->setbit(C);
    return *this;
   }

    inline const characterSetA& characterSetA::
insert (const characterA& C) const
   {guts()->setbit(C.value());
    return *this;
   }

    inline const characterSetA& characterSetA::
operator << (const characterA& C) const
   {guts()->setbit(C.value());
    return *this;
   }

    inline int characterSetA::
contains (unsigned char C) const
   {return guts()->getbit(C);}

    inline int characterSetA::
contains (const characterA& C) const
   {return guts()->getbit(C.value());}

    inline const characterSetA& characterSetA::
remove (unsigned char C) const
   {guts()->clearbit(C);
    return *this;
   }

    inline const characterSetA& characterSetA::
operator -= (unsigned char C) const
   {guts()->clearbit(C);
    return *this;
   }

    inline const characterSetA& characterSetA::
remove (const characterA& C) const
   {guts()->clearbit(C.value());
    return *this;
   }

    inline const characterSetA& characterSetA::
operator -= (const characterA& C) const
   {guts()->clearbit(C.value());
    return *this;
   }


finiteSet_RETURNING_INLINES(characterSet)


//***************************************************************************

#endif /* OATH_CHARACTERSET_H */
