#ifndef OATH_DLNODEP_H

#define OATH_DLNODEP_H

//***************************************************************************
//             OATH :: Object-oriented Abstract Type Hierarchy
//***************************************************************************
//
//  Copyright (C) 1991, 1990  Texas Instruments Incorporated
//  Permission is granted to any individual or institution
//  to use, copy, modify, and distribute this software,
//  provided that this complete copyright and permission notice
//  is maintained, intact, in all copies and supporting documentation.
//
//  Texas Instruments Incorporated provides this software "as is"
//  without express or implied warranty.
//
//***************************************************************************
//  dlNodeP
//
//  History:
//    10/90  Brian M Kennedy  Major Rewrite
//    02/90  Brian M Kennedy  Original
//
//***************************************************************************
//  dlNodeP (Doubly-Linked Node Auxiliary Type)
//
//  dlNodeP is an auxiliary structure used to implement the D-L List
//  type.  dlNodeP is NOT an Object!  dlNodeP is a class containing three
//  data items:  a const_objA, a pointer to the next dlNodeP, and
//  a pointer to the previous dlNodeP.
//
//  The following operations are defined on dlNodeP's:
//    prev()		=== Return a pointer to the previous dlNodeP.
//    thisObj()         === Return a pointer to the objG in this dlNodeP.
//    next()		=== Return a pointer to the next dlNodeP.
//    nextObj()         === Return a pointer to the objG in the next dlNodeP.
//    insertAfter(Obj)  === Make and insert a dlNodeP pointing to Obj as the
//			    next node.
//    removeNext()	=== Remove and return the next dlNodeP.
//    deleteNext()	=== Remove and delete the next dlNodeP.
//
//***************************************************************************

#include "obj.h"

/////////////////////////////////////////////////////////////////////////////
// dlNodeP Class Definition

class dlNodeP

   {public:

   	dlNodeP* Prev;
   	dlNodeP* Next;
	objA     Obj;

   	dlNodeP ()
	   :Prev(0), Next(0), Obj()
	   {}

	dlNodeP (dlNodeP* iPrev, dlNodeP* iNext, const objG* iObj)
	   :Prev(iPrev), Next(iNext), Obj(iObj)
	   {}

	    dlNodeP*
	prev ()
	   {return Prev;}

	    dlNodeP*
	next ()
	   {return Next;}

	    objA&
	thisObj ()
	   {return Obj;}

	    objA&
	nextObj ()
	   {assumed(Next, "Attempted dlNodeP::nextObj() with no Next"); 
	    return Next->Obj;
	   }

	    dlNodeP*
	insertAfter (const objG* NewObj)
	   {Next = new dlNodeP (this, Next, NewObj);
	    if(Next->Next)
		Next->Next->Prev = Next;
	    return Next;
	   }

	    dlNodeP*
	removeNext ()
	   {assumed(Next, "Attempted dlNodeP::removeNext() with no Next");
	    dlNodeP *DLN = Next;
	    Next = Next->Next;
	    if(Next)
		Next->Prev = this;
	    return DLN;
	   }

	    void
	deleteNext ()
	   {assumed(Next, "Attempted dlNodeP::deleteNext() with no Next");
	    dlNodeP *DLN = Next;
	    Next = Next->Next;
	    if(Next)
		Next->Prev = this;
	    delete DLN;
	   }

    // Dynamic Allocation
    //   for use on systems with expensive malloc

#ifdef OATH_HANDLE_NEW

	static int      AllocateCount;
	static dlNodeP *FreeList;

	    static void
	allocate ();

	    static void*
        operator new (size_t)
           {if(!FreeList)
	       allocate(); 
	    dlNodeP *DLN = FreeList;
	    FreeList = FreeList->Next;
            return DLN;
           }

            static void
        operator delete (void *DLN, size_t)
           {((dlNodeP*)DLN)->Next = FreeList;
            FreeList = (dlNodeP*)DLN;
           }

#endif

   };


//***************************************************************************

#endif  /* OATH_DLNODEP_H */
