#ifndef OATH_OATHCORE_H

#define OATH_OATHCORE_H

//***************************************************************************
//             OATH :: Object-oriented Abstract Type Hierarchy
//***************************************************************************
//
//  Copyright (C) 1991, 1990  Texas Instruments Incorporated
//  Permission is granted to any individual or institution
//  to use, copy, modify, and distribute this software,
//  provided that this complete copyright and permission notice
//  is maintained, intact, in all copies and supporting documentation.
//
//  Texas Instruments Incorporated provides this software "as is"
//  without express or implied warranty.
//
//***************************************************************************
//  oathCore (oathCoreA, oathCoreG)
//
//  History:
//    07/91  Brian M Kennedy  import, export, typeRegister
//    06/91  Brian M Kennedy  New macros & format; remove printDiagnostic
//    10/90  Brian M Kennedy  Major Rewrite
//    02/90  Brian M Kennedy  Original
//
//***************************************************************************

#include <new.h>

#include "gcConfig.h"

#include "exportP.h"

/////////////////////////////////////////////////////////////////////////////
// Runtime Assertion/Error Code

//#define NDEBUG                // turns off assumed()'s in ensure.h
                                // and environment variable checks

#include "ensure.h"

/////////////////////////////////////////////////////////////////////////////
// Standard synonyms

#ifndef FALSE
#define FALSE (0)
#endif

#ifndef TRUE
#define TRUE (!FALSE)
#endif

typedef unsigned int	// someday this should be in a standard library
	size_t;

/////////////////////////////////////////////////////////////////////////////
// Pre-defined flags

#define OATH_CONST 1
#define OATH_NIL   2
#define OATH_MARK  4
#define OATH_REF   8      // Magnitude of lsb of reference count

#define OATH_CONST_BIT 0  // Bit number of OATH_CONST
#define OATH_NIL_BIT   1  // Bit number of OATH_NIL
#define OATH_MARK_BIT  2  // Bit number of OATH_MARK
#define OATH_REF_BIT   3  // Bit number of lsb of reference count

/////////////////////////////////////////////////////////////////////////////
// INCREMENTAL_COLLECT Macro

#if (OATH_GC & OATH_GC_INC)

#define INCREMENTAL_COLLECT(Var) if(!Var->isRefd()) delete (oathCoreG*)Var

#else

#define INCREMENTAL_COLLECT(Var)

#endif

/////////////////////////////////////////////////////////////////////////////
// Runtime const check -- should be called in each non-const member function
// <<Someday this will probably be a throw.>>

#define NOT_CONST() ensure(!isConst(), "Attempted to modify a const obj!")


/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////
// oathCore Class Definitions

class oathCoreG;

/////////////////////////////////////////////////////////////////////////////
// oathCoreA Definition

class oathCoreA
   {private:
        // Members //////////
	const oathCoreG* Guts;

    public:
    	// Accessor //////////

	    const oathCoreG*
	guts () const
	   {return Guts;}

    protected:
        // Constructors //////////
	    inline
	oathCoreA ();

	    inline
	oathCoreA (const oathCoreG* iGuts);

	    inline
	oathCoreA (const oathCoreA& O);

	    inline
	~oathCoreA ();


        // Operations //////////

	    inline oathCoreA&
	operator = (const oathCoreA& O);

    public:

	    inline int
	isType (void* TypeID) const;


        // Static Operations //////////

	    static inline void
	collectGarbage (int Quick = FALSE);


	// Export Operations //////////

	    void
	export (ostream& S) const;

	    void
	export (exportP& X) const;

	    static objA
	import (istream& S);

	    static objA
	import (importP& M);

   };


/////////////////////////////////////////////////////////////////////////////
// oathCoreG Definition

class oathCoreG
   {public:
            virtual void
        vtablePosition () const;

        // Members //////////

        int      Flags;   // Reference-Count | MARK | NIL | CONST

#if (OATH_GC == OATH_GC_COM)
	oathCoreG* Prev;	  // Backward link in the list of all objs
#endif

#if (OATH_GC & OATH_GC_SAC)
        oathCoreG* Next;    // Forward link in the list of all objs
#endif

        // Constructors //////////

	    inline
        oathCoreG (int IsConst = 0);

	    inline
        oathCoreG (int, int);

	    inline virtual
        ~oathCoreG ();


        // oathCore Operations //////////

	    inline virtual int
	isType (void* TypeID) const;

	    inline virtual int
	isImplementedAs (void* TypeID) const;

	    virtual void
	export (exportP& X) const;
	    
	    static objA
	import (importP& M);

	    virtual void
	clearReferences();

	    virtual void
	setReferences();


        // oathCore Base Operations //////////

	    inline int
	isConst () const;

	    inline int
	isNil () const;

	    inline int
	is (const oathCoreG* O) const;


        // Static Operations //////////

            static void   
        collectGarbage (int Quick = FALSE);


        // Internal Operations //////////

	    inline int
	isRefd () const;

            inline void
        ref () const;    // not really const!!

            inline void
        deref () const;  // not really const!!

	    inline void
	forceConst () const; // not really const!!

	    inline int
	isMarked () const;

	    inline void
	clearMark ();

	    inline void
	setMark ();


        // Static Internal Operations //////////

	    static const char*
	registerType (const char* Name, importF Import);

            static oathCoreG* 
        nil ();

	    static void (*
	setOathNewHandler ()) ();

            static void   
        oathNewHandler ();


        // Static Members //////////

	static const char *    TypeName;

	static void         (* NextNewHandler ) ();

	static oathCoreG*      AllObjs;

        static oathCoreG*      NilPtr;

        static oathCoreG*const Nil;

   }; //oathCoreG


/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////
// oathCore Inlines

/////////////////////////////////////////////////////////////////////////////
// oathCoreG Inlines

// Constructors //////////

#if (OATH_GC == OATH_GC_COM)

    inline oathCoreG::
oathCoreG (int IsConst) // IsConst = 0
   :Flags(IsConst ? OATH_CONST : 0), Prev(0), Next(AllObjs)
   {AllObjs = this;
    if(Next) Next->Prev = this;
   }

    inline oathCoreG::
oathCoreG (int, int)
   :Flags(OATH_REF|OATH_NIL|OATH_CONST), Prev(0), Next(AllObjs)
   {AllObjs = this;
    if(Next) Next->Prev = this;
   }

    inline oathCoreG::
~oathCoreG ()
   {if(Prev) Prev->Next = Next;
    else     AllObjs = Next;
    if(Next) Next->Prev = Prev;
   }

#elif (OATH_GC == OATH_GC_SAC)

    inline oathCoreG::
oathCoreG (int IsConst) // IsConst = 0
   :Flags(IsConst ? OATH_CONST : 0), Next(AllObjs)
   {AllObjs = this;}

    inline oathCoreG::
oathCoreG (int, int)
   :Flags(OATH_REF|OATH_NIL|OATH_CONST), Next(AllObjs)
   {AllObjs = this;}

    inline oathCoreG::
~oathCoreG ()
   {}

#else

    inline oathCoreG::
oathCoreG (int IsConst) // IsConst = 0
   :Flags(IsConst ? OATH_CONST : 0)
   {}

    inline oathCoreG::
oathCoreG (int, int)
   :Flags(OATH_REF|OATH_NIL|OATH_CONST)
   {}

    inline oathCoreG::
~oathCoreG ()
   {}

#endif


// Operations //////////

    inline int oathCoreG::
isType (void*) const
   {return FALSE;}

    inline int oathCoreG::
isImplementedAs (void*) const	// There is one object implementedAs oathCoreG,
   {return FALSE;}		// Nil, but it has no "type".

    inline int oathCoreG::
isConst () const
   {return Flags & OATH_CONST;}

    inline int oathCoreG::
isNil () const
   {return Flags & OATH_NIL;}

    inline int oathCoreG::
is (const oathCoreG* O) const
   {return this == O;}


// Internal Operations //////////

    inline int oathCoreG::
isRefd () const
   {return Flags >> OATH_REF_BIT;}

    inline void oathCoreG::
ref () const   // not really const!!
   {((oathCoreG*)this)->Flags += OATH_REF;}

    inline void oathCoreG::
deref () const // not really const!!
   {((oathCoreG*)this)->Flags -= OATH_REF;}

    inline void oathCoreG::
forceConst () const // not really const!!
   {((oathCoreG*)this)->Flags |= OATH_CONST;}

    inline int oathCoreG::
isMarked () const
   {return Flags & OATH_MARK;}

    inline void oathCoreG::
clearMark ()
   {((oathCoreG*)this)->Flags &= ~OATH_MARK;}

    inline void oathCoreG::
setMark ()
   {((oathCoreG*)this)->Flags |= OATH_MARK;}


/////////////////////////////////////////////////////////////////////////////
// oathCoreA Inline Functions

// Static Operations //////////

    inline void oathCoreA::
collectGarbage (int Quick)
   {oathCoreG::collectGarbage(Quick);}


// Constructors //////////

    inline oathCoreA::
oathCoreA ()
   :Guts(oathCoreG::Nil)
   {Guts->ref();}

    inline oathCoreA::
oathCoreA (const oathCoreG* iGuts)
   :Guts(iGuts)
   {Guts->ref();}

    inline oathCoreA::
oathCoreA (const oathCoreA& O)
   :Guts(O.Guts)
   {Guts->ref();}

    inline oathCoreA::
~oathCoreA ()
   {Guts->deref();
    INCREMENTAL_COLLECT(Guts);
   }

// Operations //////////

    inline oathCoreA& oathCoreA::
operator = (const oathCoreA& O)
   {O.Guts->ref();
    Guts->deref();
    INCREMENTAL_COLLECT(Guts);
    Guts = O.Guts;
    return *this;
   }

    inline int oathCoreA::
isType (void* TypeID) const
   {return guts()->isType(TypeID);}

//***************************************************************************

#endif /* OATH_OATHCORE_H */
