#ifndef OATH_PDLQUEUE_H

#define OATH_PDLQUEUE_H

//***************************************************************************
//             OATH :: Object-oriented Abstract Type Hierarchy
//***************************************************************************
//
//  Copyright (C) 1991, 1990  Texas Instruments Incorporated
//  Permission is granted to any individual or institution
//  to use, copy, modify, and distribute this software,
//  provided that this complete copyright and permission notice
//  is maintained, intact, in all copies and supporting documentation.
//
//  Texas Instruments Incorporated provides this software "as is"
//  without express or implied warranty.
//
//***************************************************************************
//  pdlQueue (pdlQueueA, pdlQueueG)
//  pdlPos (pdlPosA, pdlPosG)
//
//  History:
//    07/91  Brian M Kennedy  import, export, typeRegister
//    06/91  Brian M Kennedy  New macros & format; remove printDiagnostic
//    10/90  Brian M Kennedy  Major Rewrite
//    02/90  Brian M Kennedy  Original
//
//***************************************************************************

#include "lifoQueue.h"

#include "slNodeP.h"

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////
// Class Definitions

class pdlPosA;
class pdlPosG;

/////////////////////////////////////////////////////////////////////////////
// class pdlQueueA

CLASS_A(pdlQueue, lifoQueue)

	// Makes //////////

            static inline pdlQueueA
        make (int MakeConst = FALSE);

            static inline pdlQueueA
        make (const seqA& S, int MakeConst = FALSE);

            static inline pdlQueueA
        make (const posA& S, const posA& B, int MakeConst = FALSE);


	// Operations //////////

	    inline pdlPosA
	makePos (int I = 0, int MakeConst = FALSE) const;

	lifoQueue_RETURNING_DEFINITIONS(pdlQueue)

   };


/////////////////////////////////////////////////////////////////////////////
// class pdlQueueG

CLASS_G(pdlQueue, lifoQueue)

    friend pdlPosG;

	// Members //////////

    	slNodeP * Header;
	pdlPosG * PosList;

	// Constructors //////////

        pdlQueueG (int IsConst = FALSE);

	    inline
	pdlQueueG (slNodeP * iHeader, int IsConst = FALSE);

	pdlQueueG (const seqG* S, int IsConst= FALSE);

	pdlQueueG (const posG* Start, const posG* Beyond, int IsConst = FALSE);

	~pdlQueueG ();

    public:
        // oathCore Operations //////////

            inline virtual int
        isImplementedAs (void* TypeId) const;

	    virtual void
	export (exportP& X) const;

	    static objA
	import (importP& M);

            virtual void
	clearReferences ();

            virtual void
	setReferences ();

        // obj Operations //////////

            virtual int
        isEqual (const objG* O) const;

            virtual objA 
        makeCopy (int MakeConst) const;

        // bag Operations //////////

            inline virtual int
        isEmpty () const;

            virtual int
        count () const;

            virtual int
        contains (const objG* O) const;

            virtual int
        containsEqual (const objG* O) const;

            virtual int
        canContain (const objG* O) const;

            virtual void
        insert (const objG* O);

            virtual void
        append (const bagG* B);

	    virtual void
	apply (void (*)(objA)) const;

	    virtual bagG*
	applyX (objA (*)(objA), bagG* B) const;

	    virtual bagG*
	applyX (bagA (*)(objA), bagG* B) const;

	    virtual bagA
	makeEmpty () const;


        // queue Operations //////////

	    inline virtual const objG*
	front () const;

	    virtual const objG*
	remove ();


        // seq Operations //////////

            inline virtual const objG*
        subscript (int I) const;

            inline virtual const objG*
        subscript (const posG* P) const;

            inline virtual posA
        makePos (int I, int MakeConst) const;

	    virtual seqA
        makeSeq (const posG* Start, const posG* Beyond, int MakeConst) const;

	    virtual seqA
        makeSeq (int Start, int Beyond, int MakeConst) const;


        // lifoQueue Operations //////////


        // pdlQueue Operations //////////

            slNodeP *
        internalPosition (int I) const;

	    inline pdlPosG*&
	posList () const;	// not really const!!!

	    void
	adjustPosList (slNodeP * CurrentPrev, slNodeP * NewPrev);

   };


/////////////////////////////////////////////////////////////////////////////
// class pdlPosA

CLASS_A(pdlPos, pos)

	pos_RETURNING_DEFINITIONS(pdlPos)

   };


/////////////////////////////////////////////////////////////////////////////
// class pdlPosG

CLASS_G(pdlPos, pos)

    friend pdlQueueG;

	// Members //////////

        const pdlQueueA List;
        slNodeP *       Prev;
	pdlPosG *       PrevPos;
	pdlPosG *       NextPos;

	// Constructors //////////

	    inline
        pdlPosG (int IsConst = FALSE);

	pdlPosG (const pdlQueueG* iList, slNodeP* iPrev, int IsConst = FALSE);

	~pdlPosG ();

    public:
        // oathCore Operations //////////

            inline virtual int
        isImplementedAs (void* TypeId) const;

	    virtual void
	export (exportP& X) const;

	    static objA
	import (importP& M);

            virtual void
	clearReferences ();

            virtual void
	setReferences ();

        // obj Operations //////////

            virtual int
        isEqual (const objG* O) const;

            inline virtual objA
        makeCopy (int MakeConst) const;

        // pos Operations //////////

            inline virtual int
        isPastEnd () const;

            virtual const objG*
        indirection () const;

            virtual void
        increment ();

            virtual void
        increment (int I);

            virtual const objG*
        find (const objG* O);

            virtual const objG*
        findEqual (const objG* O);

            virtual void
        reset (const posG* P);

            virtual void
        reset (int Index);

            inline virtual const seqG*
        parent () const;

   };


/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////
// Inline Definitions

/////////////////////////////////////////////////////////////////////////////
// pdlQueueG Inlines

// Constructors //////////

    inline pdlQueueG::
pdlQueueG (slNodeP * iHeader, int IsConst) // IsConst = FALSE)
   :lifoQueueG(IsConst), Header(iHeader), PosList(0)
   {}

// oathCore Operations //////////

    inline int pdlQueueG::
isImplementedAs (void* TypeId) const
   {return TypeId == Type;}

// bag Operations //////////

    inline int pdlQueueG::
isEmpty () const
   {return !Header->next();}

// queue Operations //////////

    inline const objG* pdlQueueG::
front () const
   {assumed(!isEmpty(), "front() attempted on an empty deq!");
    return Header->nextObj().guts();
   }


// seq Operations //////////

    inline const objG* pdlQueueG::
subscript (int I) const
   {slNodeP* N = internalPosition(I);
    assumed(N->next(), "Indexed beyond end of Seq!");
    return N->nextObj().guts();
   }

    inline const objG* pdlQueueG::
subscript (const posG* P) const
   {ensure(is(P->parent()), "Position not from this Sequence!");
    assumed(!P->isPastEnd(), "Seq[Pos] attempted with Pos that is past end!");
    return ((const pdlPosG*)P)->Prev->nextObj().guts();
   }

    inline posA pdlQueueG::
makePos (int I, int MakeConst) const
   {return new pdlPosG (this, internalPosition(I), MakeConst);}

// lifoQueue Operations //////////

// pdlQueue Operations //////////

    inline pdlPosG*& pdlQueueG::
posList () const	// not really const!!!
   {return ((pdlQueueG*)this)->PosList;}


/////////////////////////////////////////////////////////////////////////////
// pdlQueueA Inlines

INLINES(pdlQueue, lifoQueue)

    inline pdlQueueA pdlQueueA::
make (int MakeConst)
   {return new pdlQueueG(MakeConst);}

    inline pdlQueueA pdlQueueA::
make (const seqA& S, int MakeConst)
   {return new pdlQueueG(S.guts(), MakeConst);}

    inline pdlQueueA pdlQueueA::
make (const posA& S, const posA& B, int MakeConst)
   {return new pdlQueueG(S.guts(), B.guts(), MakeConst);}


    inline pdlPosA pdlQueueA::
makePos (int I, int MakeConst) const
   {return (pdlPosA&)(am().guts()->makePos(I, MakeConst));}


lifoQueue_RETURNING_INLINES(pdlQueue)


/////////////////////////////////////////////////////////////////////////////
// pdlPosG Inlines

// Constructors //////////

    inline pdlPosG::
pdlPosG (int IsConst) // IsConst = FALSE)
   :posG(IsConst)
   {}

// oathCore Operations //////////

    inline int pdlPosG::
isImplementedAs (void* TypeId) const
   {return TypeId == Type;}

// obj Operations //////////

    inline objA pdlPosG::
makeCopy (int MakeConst) const
   {return new pdlPosG(List.guts(), Prev, MakeConst);}

// pos Operations //////////

    inline int pdlPosG::
isPastEnd () const
   {return !Prev->next();}

    inline const seqG* pdlPosG::
parent () const
   {return List.guts();}


/////////////////////////////////////////////////////////////////////////////
// pdlPosA Inlines

INLINES(pdlPos, pos)

pos_RETURNING_INLINES(pdlPos)


//***************************************************************************

#endif	/* OATH_PDLQUEUE_H */
