#ifndef SIGNMAGNITUDEP_H

#define SIGNMAGNITUDEP_H

//***************************************************************************
//             OATH :: Object-oriented Abstract Type Hierarchy
//***************************************************************************
//
//  Copyright (C) 1991, 1990  Texas Instruments Incorporated
//  Permission is granted to any individual or institution
//  to use, copy, modify, and distribute this software,
//  provided that this complete copyright and permission notice
//  is maintained, intact, in all copies and supporting documentation.
//
//  Texas Instruments Incorporated provides this software "as is"
//  without express or implied warranty.
//
//***************************************************************************
//  signMagnitudeP
//
//  History:
//    08/91  Brian M Kennedy  Original
//
//***************************************************************************
//  signMagnitudeP (String Table Node)
//
//  signMagnitudeP is an auxiliary structure used to implement the numeric
//  types.  In particular, it is the standard means of communication between
//  numeric implementations.  It need not be used for the implementations 
//  themselves; however, that is typically quite efficient.
//
//  The owner of a signMagnitudeP is always the numeric that created it.
//  Further, the owner of the magnitude part of a signMagnitudeP is always
//  the numeric that created it.  A signMagnitudeP is simply a view of the
//  value of a numeric.  For efficiency, this will often be a direct look
//  at the internal structures of a numeric.  Careful!!
//
//  The basic semantics of signMagnitudeP is a view of an array of unsigned
//  longs that represent the magnitude (the zeroth element is the LSW, the
//  element at size()-1 is the MSW) and a sign.
//
//  The size should always be one or more.  The size should always be as
//  small as possible (the MSW should not be 0, unless it is the LSW).
//  The value 0 may have either sign (i.e. +0 and -0 are the same).
//
//  The following operations are defined on signMagnitudeP's:
//    zero()		--- Return non-zero if this is zero.
//    positive()	--- Return non-zero if this is positive; else zero.
//    positive(I)	--- Set this positive if I is non-zero; else negative.
//    negate()		--- If this is positive, make it negative; and v-v.
//    size()		--- Return the number of elements in the magnitude.
//    size(UI)		--- Set the number of elements in the magnitude (note
//			    this does not affect the magnitude -- the owner is
//			    responsible for maintaining the consistency).
//    incSize()		--- Increment the value of size by 1.
//    decSize()		--- Decrement the valude of size by 1.
//    incSize(UI)	--- Increment the value of size by UI.
//    decSize(UI)	--- Decrement the valude of size by UI.
//    magnitude()	--- Return the unsigned long * representing the magn.
//    magnitude(UI)	--- Return the UI'th unsigned long in the magnitude.
//
//***************************************************************************

/////////////////////////////////////////////////////////////////////////////
// signMagnitudeP Class Definition

class signMagnitudeP
   {public:
        // Members //////////

   	unsigned int    SizeAndSign;
	unsigned long * Magnitude;

        // Constructors //////////

   	signMagnitudeP ()
	   :SizeAndSign(0), Magnitude(0)
	   {}

	signMagnitudeP (const signMagnitudeP& SAM)
	   :SizeAndSign(SAM.SizeAndSign), Magnitude(SAM.Magnitude)
	   {}

	signMagnitudeP (unsigned int SAS, unsigned long * M)
	   :SizeAndSign(SAS), Magnitude(M)
	   {}

	signMagnitudeP (unsigned int Size, int Sign, unsigned long * M)
	   :SizeAndSign((Size << 1) | (Sign ? 1U : 0)), Magnitude(M)
	   {}

	// Operations //////////

	    int
	zero () const
	   {return !Magnitude[0] && (SizeAndSign <= 3);}

	    int
        positive () const
	   {return SizeAndSign & 1U;}

	    void
        positive (int P)
	   {if(P) SizeAndSign |= 1U;
	    else  SizeAndSign &= ~1U;
	   }

	    void
        negate ()
	   {SizeAndSign ^= 1U;}

	    unsigned int
        size () const
	   {return SizeAndSign >> 1;}

	    void
        size (unsigned int S)
	   {SizeAndSign &= 1U;
	    SizeAndSign |= S << 1;
	   }

	    void
        incSize ()
	   {SizeAndSign += 2U;}

	    void
        incSize (unsigned int I)
	   {SizeAndSign += I << 1;}

	    void
        decSize ()
	   {SizeAndSign -= 2U;}

	    void
        decSize (unsigned int I)
	   {SizeAndSign -= I << 1;}

	    unsigned long *
        magnitude () const
	   {return Magnitude;}

	    unsigned long *&
        magnitude ()
	   {return Magnitude;}

	    unsigned long
        magnitude (unsigned int I) const
	   {return Magnitude[I];}

	    unsigned long&
        magnitude (unsigned int I)
	   {return Magnitude[I];}

   };

/***************************************************************************/

#endif  /* SIGNMAGNITUDEP_H */
