#ifndef OATH_STRINGTABLE_H

#define OATH_STRINGTABLE_H

//***************************************************************************
//             OATH :: Object-oriented Abstract Type Hierarchy
//***************************************************************************
//
//  Copyright (C) 1991, 1990  Texas Instruments Incorporated
//  Permission is granted to any individual or institution
//  to use, copy, modify, and distribute this software,
//  provided that this complete copyright and permission notice
//  is maintained, intact, in all copies and supporting documentation.
//
//  Texas Instruments Incorporated provides this software "as is"
//  without express or implied warranty.
//
//***************************************************************************
//  stringToken (stringTokenA, stringTokenG)
//  stringTable (stringTableA, stringTableG)
//
//  History:
//    07/91  Brian M Kennedy  import, export, typeRegister
//    06/91  Brian M Kennedy  New macros & format; remove printDiagnostic
//    10/90  Brian M Kennedy  Major Rewrite
//    02/90  Brian M Kennedy  Original
//
//***************************************************************************

#include "token.h"

#include "table.h"

#include "minString.h"

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////
// Class Definitions

class stringTableA;
class stringTableG;

/////////////////////////////////////////////////////////////////////////////
// class stringTokenA

CLASS_A(stringToken, token)

	    static inline stringTokenA
	makeNewToken (const minStringA& S);

	    static inline stringTokenA
	make (const stringA& S);

	    static inline stringTokenA
	make (const char* S);

	    static inline stringTokenA
	make (const stringPosA& Start, const stringPosA& Beyond);

	    static inline stringTokenA
	make (int Uppercase, const stringA& S);

	    static inline stringTokenA
	make (int Uppercase, const char* S);

	    static inline stringTokenA
	make (int Uppercase, const stringPosA& Start, const stringPosA& Beyond);


       	    inline stringA
	string () const;

	    inline const char *
        charStar () const;

	    inline int
        hash () const;	// uses K = 4

	    inline int
        hash (int K) const;


	token_RETURNING_DEFINITIONS(stringToken)

   }; 


/////////////////////////////////////////////////////////////////////////////
// class stringTokenG

CLASS_G(stringToken, token)

    friend class stringTableG;

	// Static Members //////////

	static stringTableA StringTokens;

	// Members //////////

        minStringA String;
	int	   HashCache;

	// Constructors //////////

	    inline
        stringTokenG ();

        stringTokenG (const minStringG* IString);

        ~stringTokenG ();

    public:
        // oathCore Operations //////////

            inline virtual int
        isImplementedAs (void* TypeId) const;

	    virtual void
	export (exportP& X) const;

	    static objA
	import (importP& M);


        // stringToken Operations //////////

       	    inline const stringG*
	string () const;

	    inline const char *
        charStar () const;

	    inline int
        hash () const;	// uses K = 4

	    inline int
        hash (int K) const;

   };


/////////////////////////////////////////////////////////////////////////////
// stNode Class Definition and Inlines

#include "stNodeP.h"

/////////////////////////////////////////////////////////////////////////////
// class stringTableA

CLASS_A(stringTable, table)

	    static inline stringTableA
	make (unsigned int Buckets = 255, int MakeConst = FALSE);


	    inline stringTokenA
	lookup (const stringA& S) const;

	    inline stringTokenA
	lookup (const char * C) const;

	    inline stringTokenA
	lookup (const stringPosA& Start, const stringPosA& Beyond) const;

	    inline stringTokenA
	lookup (int Uppercase, const stringA& S) const;

	    inline stringTokenA
	lookup (int Uppercase, const char * C) const;

	    inline stringTokenA
	lookup (int Uppercase,
	        const stringPosA& Start, const stringPosA& Beyond) const;


	table_RETURNING_DEFINITIONS(stringTable)

	// Leafing //////////

	    inline objA
	assoc (const stringTokenA& ST);

   };


/////////////////////////////////////////////////////////////////////////////
// class stringTableG

CLASS_G(stringTable, table)

    friend class stringTokenG;

	// Members //////////

        stNodeP**    Table;
	unsigned int Count;
	unsigned int Buckets;

	// Constructors //////////

        stringTableG (unsigned int IBuckets = 1, int IsConst = FALSE);

    public:
        // oathCore Operations //////////

            inline virtual int
        isImplementedAs (void* TypeId) const;

	    virtual void
	export (exportP& X) const;

	    static objA
	import (importP& M);


        // obj Operations //////////

            virtual
        isEqual (const objG* O) const;

            virtual objA
        makeCopy (int MakeConst) const;


        // bag Operations //////////

            inline virtual int
        isEmpty () const;

            inline virtual int
        count () const;

            virtual int
        contains (const objG* O) const;

            virtual int
        containsEqual (const objG* O) const;

            virtual int
        canContain (const objG* O) const;

            virtual void
        insert (const objG* O);

            virtual void
        append (const bagG* B);

	    virtual void
	apply (void (*)(objA)) const;

	    virtual bagG*
	applyX (objA (*)(objA), bagG* B) const;

	    virtual bagG*
	applyX (bagA (*)(objA), bagG* B) const;

	    virtual bagA
	makeEmpty () const;


        // table Operations //////////

	    virtual void
	insertAssoc (const objG* O, const objG* A);

	    virtual void
	reassoc (const objG* O, const objG* A);

	    virtual const objG*
	remove (const objG* O);

	    virtual const objG*
	assoc (const objG* O) const;


        // stringTable Operations //////////

	    inline const stringTokenG*
	lookup (const stringG* S) const;

	    inline const stringTokenG*
	lookup (const char * C) const;

	    inline const stringTokenG*
	lookup (const stringPosG* Start, const stringPosG* Beyond) const;

	    inline const stringTokenG*
	lookup (int Uppercase, const stringG* S) const;

	    inline const stringTokenG*
	lookup (int Uppercase, const char * C) const;

	    inline const stringTokenG*
	lookup (int Uppercase,
	        const stringPosG* Start, const stringPosG* Beyond) const;

   };


/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////
// Inline Definitions

/////////////////////////////////////////////////////////////////////////////
// stringTokenG Inlines

// Constructors //////////

    inline stringTokenG::
stringTokenG ()
   :tokenG(), String(minStringG::Nil)
   {}


// oathCore Operations //////////

    inline int stringTokenG::
isImplementedAs (void* TypeId) const
   {return TypeId == Type;}


// stringToken Operations //////////

    inline const stringG* stringTokenG::
string () const
   {return String.guts();}

    inline const char * stringTokenG::
charStar () const
   {return String.charStar();}

    inline int stringTokenG::
hash () const	// uses K = 4
   {return HashCache;}

    inline int stringTokenG::
hash (int K) const
   {return String.hash(K);}


/////////////////////////////////////////////////////////////////////////////
// stringTokenA Inlines

INLINES(stringToken, token)

// stringToken Operations //////////

    inline stringA stringTokenA::
string () const
   {return am().guts()->string();}

    inline const char * stringTokenA::
charStar () const
   {return am().guts()->charStar();}

    inline int stringTokenA::
hash () const	// uses K = 4
   {return am().guts()->hash();}

    inline int stringTokenA::
hash (int K) const
   {return am().guts()->hash(K);}


token_RETURNING_INLINES(stringToken)


/////////////////////////////////////////////////////////////////////////////
// stringTableG Inlines

// oathCore Operations //////////

    inline int stringTableG::
isImplementedAs (void* TypeId) const
   {return TypeId == Type;}


// bag Operations //////////

    inline int stringTableG::
isEmpty () const
   {return !Count;}

    inline int stringTableG::
count () const
   {return Count;}


// stringTable Operations //////////

    inline const stringTokenG* stringTableG::
lookup (const stringG* S) const
   {stNodeP* Bucket = Table[S->hash() % Buckets];
    return Bucket ? Bucket->lookup(S) : stringTokenG::Nil;
   }

    inline const stringTokenG* stringTableG::
lookup (const char * C) const
   {stNodeP* Bucket = Table[stringG::hash(C) % Buckets];
    return Bucket ? Bucket->lookup(C) : stringTokenG::Nil;
   }

    inline const stringTokenG* stringTableG::
lookup (const stringPosG* Start, const stringPosG* Beyond) const
   {stNodeP* Bucket = Table[stringG::hash(Start, Beyond)%Buckets];
    return Bucket ? Bucket->lookup(Start, Beyond) : stringTokenG::Nil;
   }

    inline const stringTokenG* stringTableG::
lookup (int Uppercase, const stringG* S) const
   {int HashValue = Uppercase ? S->hashUpper() : S->hashLower();
    stNodeP* Bucket = Table[HashValue % Buckets];
    return Bucket ? Bucket->lookupAnyCase(S) : stringTokenG::Nil;
   }

    inline const stringTokenG* stringTableG::
lookup (int Uppercase, const char * C) const
   {int HashValue = Uppercase ? stringG::hashUpper(C) : stringG::hashLower(C);
    stNodeP* Bucket = Table[HashValue % Buckets];
    return Bucket ? Bucket->lookupAnyCase(C) : stringTokenG::Nil;
   }

    inline const stringTokenG* stringTableG::
lookup (int Uppercase, const stringPosG* Start, const stringPosG* Beyond) const
   {int HashValue = Uppercase ? stringG::hashUpper(Start, Beyond)
			      : stringG::hashLower(Start, Beyond);
    stNodeP* Bucket = Table[HashValue % Buckets];
    return Bucket ? Bucket->lookupAnyCase(Start, Beyond) : stringTokenG::Nil;
   }


/////////////////////////////////////////////////////////////////////////////
// stringTableA Inlines

INLINES(stringTable, table)

    inline stringTableA stringTableA::
make (unsigned int Buckets, int MakeConst)
   {return new stringTableG (Buckets, MakeConst);}


    inline stringTokenA stringTableA::
lookup (const stringA& S) const
   {return am().guts()->lookup(S.guts());}

    inline stringTokenA stringTableA::
lookup (const char * C) const
   {return am().guts()->lookup(C);}

    inline stringTokenA stringTableA::
lookup (const stringPosA& Start, const stringPosA& Beyond) const
   {return am().guts()->lookup(Start.guts(), Beyond.guts());}

    inline stringTokenA stringTableA::
lookup (int Uppercase, const stringA& S) const
   {return am().guts()->lookup(Uppercase, S.guts());}

    inline stringTokenA stringTableA::
lookup (int Uppercase, const char * C) const
   {return am().guts()->lookup(Uppercase, C);}

    inline stringTokenA stringTableA::
lookup (int Uppercase, const stringPosA& Start, const stringPosA& Beyond) const
   {return am().guts()->lookup(Uppercase, Start.guts(), Beyond.guts());}


table_RETURNING_INLINES(stringTable)

    inline objA stringTableA::
assoc (const stringTokenA& ST)
   {return am().guts()->Table[ST.hash() % guts()->Buckets]
	              ->assoc(ST.guts());
   }


/////////////////////////////////////////////////////////////////////////////
// stringTokenA Inlines continued (makes that use stringTable::lookup())

    inline stringTokenA stringTokenA::
makeNewToken (const minStringA& S)
   {return new stringTokenG (S.guts());}

    stringTokenA stringTokenA::
make (const stringA& S)
   {stringTokenA ST = stringTokenG::StringTokens.lookup(S);
    if(!ST) ST = makeNewToken(minStringA::make(S, TRUE));
    return ST;
   }

    inline stringTokenA stringTokenA::
make (const char* S)
   {stringTokenA ST = stringTokenG::StringTokens.lookup(S);
    if(!ST) ST = makeNewToken(minStringA::make(S, TRUE));
    return ST;
   }

    inline stringTokenA stringTokenA::
make (const stringPosA& Start, const stringPosA& Beyond)
   {stringTokenA ST = stringTokenG::StringTokens.lookup(Start, Beyond);
    if(!ST) ST = makeNewToken(minStringA::make(Start, Beyond, TRUE));
    return ST;
   }

    inline stringTokenA stringTokenA::
make (int Uppercase, const stringA& S)
   {stringTokenA ST = stringTokenG::StringTokens.lookup(Uppercase, S);
    if(!ST) ST = makeNewToken(minStringA::make(Uppercase, S, TRUE));
    return ST;
   }

    inline stringTokenA stringTokenA::
make (int Uppercase, const char* S)
   {stringTokenA ST = stringTokenG::StringTokens.lookup(Uppercase, S);
    if(!ST) ST = makeNewToken(minStringA::make(Uppercase, S, TRUE));
    return ST;
   }

    inline stringTokenA stringTokenA::
make (int Uppercase, const stringPosA& Start, const stringPosA& Beyond)
   {stringTokenA ST = stringTokenG::StringTokens.lookup(Uppercase, Start, 
								   Beyond);
    if(!ST) ST = makeNewToken(minStringA::make(Uppercase, Start, Beyond, TRUE));
    return ST;
   }


//***************************************************************************

#endif /* OATH_STRINGTABLE_H */
